use std::rc::Rc;

use crate::{
    Builtin, DeBrujinAst,
    exec::builtins::DeBrujinBuiltInAst,
    parse::Constant,
    types::{PrimitiveType, TaggedType, Type, TypeTag},
};

pub struct OpCond;
struct OpCond1(bool);
struct OpCond2(bool, DeBrujinBuiltInAst);

impl Builtin for OpCond {
    fn name(&self) -> String {
        "if".to_string()
    }

    fn r#type(&self) -> TaggedType {
        TaggedType::Tagged(
            TypeTag::Any,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                PrimitiveType::Bool,
                Type::arrow("a", Type::arrow("a", "a")),
            ))),
        )
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        match rhs {
            DeBrujinBuiltInAst::Constant(Constant::Bool(b)) => {
                Some(DeBrujinBuiltInAst::Builtin(Rc::new(OpCond1(b))))
            }
            _ => None,
        }
    }
}

impl Builtin for OpCond1 {
    fn name(&self) -> String {
        format!("if{}1", self.0)
    }

    fn r#type(&self) -> TaggedType {
        TaggedType::Tagged(
            TypeTag::Any,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a"),
            ))),
        )
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::FreeVariable("if".to_string())),
            Box::new(DeBrujinAst::Constant(Constant::Bool(self.0))),
        )
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        Some(DeBrujinBuiltInAst::Builtin(Rc::new(OpCond2(self.0, rhs))))
    }
}

impl Builtin for OpCond2 {
    fn name(&self) -> String {
        format!("if{}2", self.0)
    }

    fn r#type(&self) -> TaggedType {
        TaggedType::Tagged(
            TypeTag::Any,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow("a", "a"))),
        )
    }

    fn to_ast(&self) -> DeBrujinAst {
        DeBrujinAst::Application(
            Box::new(DeBrujinAst::Application(
                Box::new(DeBrujinAst::FreeVariable("if".to_string())),
                Box::new(DeBrujinAst::Constant(Constant::Bool(self.0))),
            )),
            Box::new(self.1.clone().into()),
        )
    }

    fn apply(&self, rhs: DeBrujinBuiltInAst) -> Option<DeBrujinBuiltInAst> {
        Some(if self.0 { self.1.clone() } else { rhs })
    }
}
