use std::mem;

pub struct VecMap<K: Eq, V> {
    map: Vec<(K, V)>,
}

impl<K: Eq, V> VecMap<K, V> {
    pub fn new() -> Self {
        Self { map: Vec::new() }
    }

    pub fn insert(&mut self, key: K, val: V) -> Option<(K, V)> {
        if let Some(entry) = self.map.iter_mut().find(|(k, _)| *k == key) {
            Some(mem::replace(entry, (key, val)))
        } else {
            self.map.push((key, val));
            None
        }
    }

    pub fn get(&self, key: &K) -> Option<&V> {
        self.map.iter().find_map(|(k, v)| (k == key).then_some(v))
    }

    pub fn map_keys<F: Fn(&mut K)>(&mut self, f: F) {
        self.map.iter_mut().for_each(|(k, _)| f(k));
    }
}

impl<K: Eq, V> Default for VecMap<K, V> {
    fn default() -> Self {
        Self {
            map: Vec::default(),
        }
    }
}

impl<K: Eq + Clone, V: Clone> Clone for VecMap<K, V> {
    fn clone(&self) -> Self {
        Self {
            map: self.map.clone(),
        }
    }
}
